#define __NEW_STARLET 1
#pragma module PCA "X01-00"

/*
**++
**
** Copyright (c) 2014 James F. Duff.  ALL RIGHTS RESERVED.
**
** License Terms and Conditions:
**
** You may freely copy and distribute verbatim copies of this software package
** as you receive it, in any medium, provided that you conspicuously and
** appropriately publish on each copy this copyright notice and
** disclaimer of warranty; keep intact all the notices that refer to these
** license terms and conditions and to the absence of any warranty; and give
** any other recipients of the software package a copy of these license
** terms and conditions along with the software.
**
** You may modify the software for your own use, but you MAY NOT distribute
** modified copies of the software.
**
** By copying or distributing the software package you indicate your
** acceptance of this license to do so, and all its terms and conditions.
**
** Each time you redistribute the software package the recipient
** automatically receives a license from the original licensor to copy,
** distribute or modify the software package subject to these terms and
** conditions. You may not impose any further restrictions on the recipients'
** exercise of the rights granted herein.
**
** BECAUSE THE SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
** FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
** OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDER PROVIDE THE SOFTWARE
** "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,
** INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
** AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY
** AND PERFORMANCE OF THE SOFTWARE IS WITH YOU.  SHOULD THE SOFTWARE PROVE
** DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
** CORRECTION.
**
** YOU WARRANT THAT IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO
** YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR
** CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE
** SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF THE
** SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE, FIRMWARE OR HARDWARE), EVEN IF
** SUCH HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
**
** NO TITLE TO AND OWNERSHIP OF THE SOFTWARE PACKAGE IS HEREBY TRANSFERRED.
**
** Send suggestions, bug reports, money ;-) to jim@eight-cubed.com
**
** Now that's out of the way, on with the code...
**
**--
*/

/*
**++
**
**  MODULE DESCRIPTION:
**
**	This module implements a poor man's version of the program counter
**	sampling available in the DECset PCA product.
**
**	To do this, the module uses LIB$INITIALIZE (in module PCA_INIT.MAR) to
**	register PCA_ARM as an initialization routine.  This routine is called
**	after the debugger (if there is one) and before control passes to the
**	main routine.  See the "OpenVMS Calling Standard" for additional
**	information on how this all works.
**
**	Practically, you can link PCA.OBJ and PCA_INIT.OBJ against the existing
**	object(s) you wish to have PC information collected from.  No need to
**	modify the analysed code.
**
**	When you run the code with PCA and PCA_INIT linked in, PC data will be
**	collected in a file located in the default directory named
**	filename.PCC, where filename is the filename of the executable.
**
**	Note that this collecter will only collect information where the
**	symbolic image name of the PC matches the name of the executable.
**	So, for example, if you have constructed TEST.EXE like so:
**
**		$ cc module1.c
**		$ cc module2.c
**		$ cc sharable.c
**		$ link/share sharable.c
**		$ link/exe=test.exe module1,module2,pca,pca_init,sys$input/opt
**		sharable.exe/share
**		^Z
**
**	PC data will be collected for PCs that fall in the range of code
**	contributed by module1.c and module2.c.  But code contributed by
**	sharable.c will not be reported on.  To collect information about a
**	sharable image, temporarily hard link it into the executable.
**
**	Once PC information has been collected, you can dump it from the PCC
**	file using the PCA_DUMP.C code.
**
** RESTRICTIONS:
**
**	The timer routine that executes to record the PC only updates
**	the PCC file every 10 seconds.  This reduces the I/O this
**	routine does, limiting the performance impact the collector
**	has on the program under analysis.  However, it also implies
**	that you won't get meaningful data on short running programs.
**	(Why you'd analyse a short running program escapes me anyway.)
**
**	The output file is limited to 64 PC values.  This should be
**	more than enough PCs to localise a performance botteneck.
**	You can expand the number of slots in the section by modifying
**	values in PCA.H if you really need to.
**
**	This program uses a user mode AST to collect the data.  If you
**	have a program that does significant work at user mode AST
**	level or uses elevated modes (supervisor, executive, kernel),
**	this program will be useless to you. 
**
**	Any errors in PCA will cause the main program to exit.
**
**	This is currenly IA64 only code.  I might back-port to Alpha it in my
**	copious free time.
**
**  AUTHORS:
**
**      James F. Duff
**
**  CREATION DATE:
**
**      17-Mar-2014
**
**  MODIFICATION HISTORY:
**
**      X01-00      Jim Duff        17-Mar-2014
**      Original version of module.
**
**--
*/

#include <stdio.h>
#include <stdlib.h>
#include <ssdef.h>
#include <stsdef.h>
#include <descrip.h>
#include <string.h>
#include <libicb.h>
#include <efndef.h>
#include <gen64def.h>
#include <va_rangedef.h>
#include <rms.h>
#include <iosbdef.h>
#include <secdef.h>
#include <jpidef.h>
#include <iledef.h>
#include <fscndef.h>
#include <tbkdef.h>
#include <tbk$routines.h>
#include <lib$routines.h>
#include <starlet.h>

#include "pca.h"

#define errchk_stop(arg) if (!$VMS_STATUS_SUCCESS(arg)) (void)lib$stop(arg);

/*
** Forward def
*/
static void pca_get_pc (void);


/******************************************************************************/
extern void PCA_ARM (void) {
/*
** Set a 1 second timer to execute routine pca_get_pc.
*/

static GENERIC_64 delay;

static int r0_status;
static int first_time = TRUE;

static $DESCRIPTOR (delay_d, "0 00:00:01.00");

    if (first_time) {
	/*
	** Convert our delay to binary.
	*/
	first_time = FALSE;
	r0_status = sys$bintim (&delay_d,
				&delay);
	errchk_stop (r0_status);
    }

    /*
    ** Set the timer to expire and execute the pca_get_pc routine.
    */
    r0_status = sys$setimr (EFN$C_ENF,
			    &delay,
			    &pca_get_pc,
			    0,
			    0);
    errchk_stop (r0_status);
}


/******************************************************************************/
static void pca_get_pc (void) {
/*
** Execute this routine as an AST and collect PC information.
** Once finished, re-arm the timer to execute us again in 1 second.
*/

static INVO_CONTEXT_BLK icb;

static struct FAB fab;

static TBK_API_PARAM tbkitms;

static IOSB iosb;

static VA_RANGE in_addr;
static VA_RANGE out_addr;

static ILE2 fsitms[] = { 0, FSCN$_NAME, NULL,
                         0, 0, NULL };

static unsigned __int64 flags = 0;

static struct sec_s *sec;

static int r0_status;
static int finished;
static int first_time = TRUE;
static int i;
static int low_slot;
static int processed;
static int jpiitm = JPI$_IMAGNAME;

static const unsigned int prot = 0x00FF;
static unsigned int line;
static unsigned int min;
static unsigned int exe_cnt;

static char outfile[NAM$C_MAXRSS];
static char main_image[NAM$C_MAXRSS];
static char image[512];
static char module[512];
static char routine[512];

static struct dsc$descriptor_s outfile_d = { 0,
					     DSC$K_DTYPE_T,
					     DSC$K_CLASS_S,
					     outfile };
static struct dsc$descriptor_s main_image_d = { NAM$C_MAXRSS,
					        DSC$K_DTYPE_T,
					        DSC$K_CLASS_S,
					        main_image };
static struct dsc$descriptor_s image_d = { sizeof (image) - 1,
                                            DSC$K_DTYPE_T,
                                            DSC$K_CLASS_S,
                                            image };
static struct dsc$descriptor_s module_d = { sizeof (module) - 1,
                                             DSC$K_DTYPE_T,
                                             DSC$K_CLASS_S,
                                             module };
static struct dsc$descriptor_s routine_d = { sizeof (routine) - 1,
                                              DSC$K_DTYPE_T,
                                              DSC$K_CLASS_S,
                                              routine };
static $DESCRIPTOR (section_name_d, "PCDUMP_SECTION");


    if (first_time) {
	first_time = FALSE;

	/*
	** Do some initialization.
	** Initialize the traceback block so we can do symbolization.
	*/
	(void)memset (&tbkitms, 0, sizeof (TBK_API_PARAM));
	tbkitms.tbk$w_length = TBK$K_LENGTH;
	tbkitms.tbk$b_version = TBK$K_VERSION;
	tbkitms.tbk$pq_image_desc = (struct _descriptor *)&image_d;
	tbkitms.tbk$pq_module_desc = (struct _descriptor *)&module_d;
	tbkitms.tbk$pq_routine_desc = (struct _descriptor *)&routine_d;
	tbkitms.tbk$pq_listing_lineno = &line;
	tbkitms.tbk$pq_symbolize_flags = &flags;

	/*
	** Find the name of our image.
	*/
	r0_status = lib$getjpi (&jpiitm,
				0,
				0,
				0,
				&main_image_d,
				&main_image_d.dsc$w_length);
	errchk_stop (r0_status);

	/*
	** Find the filename of our image.
	*/
	r0_status = sys$filescan (&main_image_d,
		                  fsitms,
				  0,
				  0,
                                  0);
	errchk_stop (r0_status);

	/*
	** Format an output filename.
	*/
	outfile_d.dsc$w_length = sprintf (outfile,
					  "%-.*s.PCC",
					  fsitms[0].ile2$w_length,
					  (char *)fsitms[0].ile2$ps_bufaddr);

	/*
	** Initialize out output file's FAB.  Note UFO so we can map it
	** with a section.
	*/
	fab = cc$rms_fab;
	fab.fab$b_fns = outfile_d.dsc$w_length;
	fab.fab$l_fna = outfile_d.dsc$a_pointer;
	fab.fab$l_alq = SECTION_SIZE;
	fab.fab$l_fop = FAB$M_UFO|FAB$M_CBT;
	fab.fab$b_fac = FAB$M_PUT;
	fab.fab$b_rtv = -1;

	/*
	** Create our output file.  
	*/
	r0_status = sys$create (&fab);
	errchk_stop (r0_status);

	/*
	** Map a section on our output file for write.
	*/
	r0_status = sys$crmpsc (&in_addr,
                                &out_addr,
                                0,
                                SEC$M_EXPREG|SEC$M_WRT,
                                &section_name_d,
                                0,
                                0,
                                fab.fab$l_stv,
                                SECTION_SIZE,
                                0,
                                prot,
                                0);
	errchk_stop (r0_status);

	/*
	** Let's make it easy to access via C.
	*/
	sec = out_addr.va_range$ps_start_va;

	/*
	** Initialize our section to all zeros and write those zeros to the
	** file.
	*/
	(void)memset (sec, 0, SECTION_SIZE_BYTES);
        r0_status = sys$updsecw (&out_addr,
             			 0,
                             	 0,
                             	 1,
                              	 EFN$C_ENF,
                             	 &iosb,
                             	 0,
                             	 0);
	errchk_stop (r0_status);
	errchk_stop (iosb.iosb$w_status);
    }

    /*
    ** Initialize an invokation context block.
    */
    (void)memset (&icb, 0, sizeof (INVO_CONTEXT_BLK));
    r0_status = lib$i64_init_invo_context (&icb, LIBICB$K_INVO_CONTEXT_VERSION);
    errchk_stop (r0_status);

    /*
    ** Cache context data to make the stack walk efficient.
    */
    icb.libicb$v_uo_flag_cache_unwind = 1;

    /*
    ** Retrieve the current invokation context.  This will describe the call
    ** frame of this routine (pca_pc_ast).  We're not interested in this, but
    ** where we will return to eventually in the main image code.
    */
    lib$i64_get_curr_invo_context (&icb);

    finished = FALSE;
    while (!finished) {
	/*
	** Now walk the call frames by getting the next one on the stack.
	*/
        r0_status = lib$i64_get_prev_invo_context (&icb);
	if (!$VMS_STATUS_SUCCESS (r0_status)) {
            finished = TRUE;
        } else {
            if (icb.libicb$v_bottom_of_stack) {
                finished = TRUE;
            }
        }
	if (finished) {
	    /*
	    ** If we hit the bottom of the stack, end the loop.
	    */
	    continue;
	}

	if ((unsigned __int64)icb.libicb$ih_pc <
	    (unsigned __int64)0xffffffff7fffffff) {
	    /*
	    ** If we have a PC not in system space, symbolize it.
	    */
	    tbkitms.tbk$q_faulting_pc = (unsigned __int64)icb.libicb$ih_pc;
	    r0_status = tbk$i64_symbolize (&tbkitms);
	    if (!$VMS_STATUS_SUCCESS (r0_status)) {
		/*
		** Something went wrong.  Just skip this frame.
		*/
		continue;
	    }

	    /*
	    ** Is it in our image?
	    */
	    if (memcmp (image,
			fsitms[0].ile2$ps_bufaddr,
			fsitms[0].ile2$w_length) == 0) {

		/*
		** Look through the slots in our output file looking for either
		** an unused one, one that already has this PC, or a low count
		** one that we can replace.
		*/
	        min = 99999999;
	        processed = FALSE;
	        for (i = 0; i < SLOT_COUNT; i++) {
		    if (sec->slot[i].count < min) {
		        low_slot = i;
		        min = sec->slot[i].count;
		    }
		    if (sec->slot[i].count == 0) {
		        sec->slot[i].count = 1;
		        sec->slot[i].pc = icb.libicb$ih_pc;
		        processed = TRUE;
		        break;
		    }
		    if (sec->slot[i].pc == icb.libicb$ih_pc) {
		        sec->slot[i].count++;
		        processed = TRUE;
		        break;
		    }
	        }
	        if (!processed) {
		    sec->slot[low_slot].count = 1;
		    sec->slot[low_slot].pc = icb.libicb$ih_pc;
	        }

	        if ((++exe_cnt % 10) == 0) {
		    /*
		    ** Write out the updated section every 10 times through
		    ** this routine.
		    */
	            r0_status = sys$updsecw (&out_addr,
                             	             0,
                             	             0,
                             	             1,
                              	             EFN$C_ENF,
                             	             &iosb,
                             	             0,
                             	             0);
	            errchk_stop (r0_status);
	            errchk_stop (iosb.iosb$w_status);
	        }

		/*
		** Set the finished flag so we don't got further down the
		** stack once we've seen an appropriate PC.
		*/
	        finished = TRUE;
	    }
	}
    }

    /*
    ** Free any cached context data.
    */
    (void)lib$i64_prev_invo_end (&icb);

    /*
    ** Set the timer to expire again.
    */
    PCA_ARM ();
}
